#define USEFPYROM
/***********************************************************************
*
* simfpy.c - FD800 floppy IO processing for the TI 990 Simulator.
*
* Changes:
*   05/29/03   DGP   Original.
*   06/20/03   DGP   Added interrupt support.
*   07/01/03   DGP   Fixed floppy unit on STCR of status.
*   07/08/03   DGP   Changed FLOPPY to return status if unit not present.
*   11/06/03   DGP   Make a seperate module.
*   01/26/04   DGP   Added memory size option.
*
* The disk is formatted as follows:
*    N bytes: raw data for each sector - the order is:
*      track 0 sector 0
*      track 0 sector 1, ...
*      track 0 sector N,
*      track 1 sector 0, ...
*      track M sector N, ...
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>

#if defined (_WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#endif

#if defined(UNIX)
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"

extern uint16 pcreg;	/* The program PC */
extern uint16 statreg;	/* The program status register */
extern uint16 wpreg;	/* The program Workspace Pointer */
extern uint16 lights;	/* The panel lights */

extern int run;
extern int idle;
extern int fault;
extern int devcnt;
extern int model;
extern int mapenabled;
extern int pancount;
extern uint32 memlen;
extern char view[MAXVIEW][81];

extern uint8 memory[MEMSIZE];
extern Device devices[MAXDEVICES];


/***********************************************************************
* fpyseto - Set CRU bit to one.
***********************************************************************/

void
fpyseto (Device *dev, uint8 disp)
{
   uint16 devaddr;

   devaddr = R12;

#ifdef DEBUGFLOPPY
   fprintf (stderr, "INTMASK %X CRU %04X SBO %d\n", GET_MASK, devaddr, disp);
#endif
   return;
}

/***********************************************************************
* fpysetz - Set CRU bit to zero.
***********************************************************************/

void
fpysetz (Device *dev, uint8 disp)
{
   uint16 devaddr;

   devaddr = R12;

#ifdef DEBUGFLOPPY
   fprintf (stderr, "INTMASK %X CRU %04X SBZ %d\n", GET_MASK, devaddr, disp);
#endif
   return;
}

/***********************************************************************
* fpytb - Test CRU bit.
***********************************************************************/

void
fpytb (Device *dev, uint8 disp)
{
   uint16 devaddr;

   devaddr = R12;
   
#ifdef DEBUGFLOPPY
   fprintf (stderr, "INTMASK %X CRU %04X TB %d\n", GET_MASK, devaddr, disp);
#endif

   if (devaddr == dev->devaddr) /* Data Transfer */
   {
      disp -= 16;
   }

   switch (disp)
   {
   case 0: /* OP complete (OPCOMP) */
   case 1: /* Xfer ready (CXFER) */
   case 15: /* Interrupt */
      SET_EQ;
      break;

   case 10: /* Controller busy (CBUSY) */
   default:
      CLR_EQ;
   }
   return;
}

/***********************************************************************
* fpyldcr - Load CRU.
***********************************************************************/

void
fpyldcr (Device *dev, uint16 cnt, uint16 dat)
{
   int i;
   int genint;
   uint16 devaddr;
#ifdef DEBUGFLOPPY
   uint8 ch;
   uint8 ch1;
#endif

   devaddr = R12;

#ifdef DEBUGFLOPPY
   fprintf (stderr, "INTMASK %X LDCR %04X", GET_MASK, devaddr);
#endif

   genint = dev->intenabled;
   if (devaddr == dev->devaddr) /* Data transfer */
   {
      for (i = 0; i < devcnt; i++) /* find selected unit */
	 if (devices[i].select && devices[i].devaddr == devaddr)
	    break;
      if (i == devcnt)
      {
	 dev->info.fpyinfo.laststat = 0x4; /* Drive not ready */
	 return;
      }
      dev = &devices[i];
#ifdef DEBUGFLOPPY
      ch = (dat >> 8) & 0xFF;
      if (!isprint (ch)) ch = '.';
      ch1 = dat & 0xFF;
      if (!isprint (ch1)) ch1 = '.';
      fprintf (stderr, " DATA %04X %c%c\n", dat, ch, ch1);
#endif
      fputc ((dat >> 8) & 0xFF, dev->fd);
      fputc (dat & 0xFF, dev->fd);
      dev->count -= 2;
      if (dev->count && genint)
	 geninterrupt (dev->intlvl);
   }

   else /* Command */
   {
      int diskloc;
      uint16 cmd;
      uint16 unit;
      uint16 sec;
      uint16 trk;

      cmd = dat;
#ifdef DEBUGFLOPPY
      fprintf (stderr, " CMD %04X", cmd);
#endif

      /* If Clear Status port, return */
      if (((cmd & 0xF000) >> 12) == 15)
      {
#ifdef DEBUGFLOPPY
	 fprintf (stderr, "\n");
#endif
	 for (i = 0; i < devcnt; i++) /* find selected unit */
	    if (devices[i].cdevaddr == devaddr)
	    {
	       dev->cbreak = FALSE;
	       dev->select = FALSE;
	       dev->info.fpyinfo.laststat = 0;
	    }
	 return;
      }

      unit = (cmd & 0x0C00) >> 10;
#ifdef DEBUGFLOPPY
      fprintf (stderr, " UNIT %d", unit);
#endif

      for (i = 0; i < devcnt; i++) /* find selected unit */
	 if (devices[i].cdevaddr == devaddr && devices[i].unit == unit)
	    break;
      if (i == devcnt)
      {
#ifdef DEBUGFLOPPY
	 fprintf (stderr, " not selected\n");
#endif
	 dev->info.fpyinfo.laststat = 0x4 | (unit << 13); /* Drive not ready */
	 if (genint)
	    geninterrupt (dev->intlvl);
	 return;
      }
      dev = &devices[i];

      switch ((cmd & 0xF000) >> 12)
      {
      case 0: /* Select */
	 dev->select = TRUE;
	 break;

      case 14: /* IPL */
         dev->count = FPYSECTLEN;
      case 1: /* Seek */
	 trk = cmd & 0x007F;
	 diskloc = (FPYSECTLEN * FPYSECTTRK) * trk;
	 if (fseek (dev->fd, diskloc, SEEK_SET) < 0)
	 {
	    sprintf (view[0], "fpyldcr: seek failed: %s",
		  strerror (ERRNO));
	    sprintf (view[1], "filename: %s", dev->file);
	    run = FALSE;
	    return;
	 }
#ifdef DEBUGFLOPPY
	 fprintf (stderr, " TRACK %d LOC %ld", trk, diskloc);
#endif
	 dev->info.fpyinfo.track = diskloc;
	 dev->select = TRUE;
	 break;

      case 2: /* Restore */
	 dev->select = TRUE;
	 dev->info.fpyinfo.track = 0;
	 break;

      case 3: /* Sector Length */
	 dev->select = TRUE;
	 break;

      case 4: /* Read */
      case 6: /* Read Unformatted */
      case 7: /* Write */
      case 8: /* Write Delete */
	 sec = cmd & 0x001F;
	 diskloc = dev->info.fpyinfo.track + (FPYSECTLEN * (sec-1));
#ifdef DEBUGFLOPPY
	 fprintf (stderr, " SECTOR %d LOC %ld", sec, diskloc);
#endif
	 if (fseek (dev->fd, diskloc, SEEK_SET) < 0)
	 {
	    sprintf (view[0], "fpyldcr: seek failed: %s",
		  strerror (ERRNO));
	    sprintf (view[1], "filename: %s", dev->file);
	    run = FALSE;
	    return;
	 }
         dev->count = FPYSECTLEN;
	 dev->select = TRUE;
	 break;

      case 5: /* Read ID */
	 break;

      case 9: /* Format Track */
	 trk = cmd & 0x007F;
	 diskloc = (FPYSECTLEN * FPYSECTTRK) * trk;
#ifdef DEBUGFLOPPY
	 fprintf (stderr, " TRACK %d LOC %ld", trk, diskloc);
#endif
	 if (fseek (dev->fd, diskloc, SEEK_SET) < 0)
	 {
	    sprintf (view[0], "fpyldcr: seek failed: %s",
		  strerror (ERRNO));
	    sprintf (view[1], "filename: %s", dev->file);
	    run = FALSE;
	    return;
	 }
	 for (i = 0; i < FPYSECTTRK*FPYSECTLEN; i++) fputc ('\0', dev->fd);
	 dev->info.fpyinfo.track = trk+1;
	 dev->select = TRUE;
	 break;

      case 10: /* Load Int Mask */
	 dev->select = TRUE;
	 dev->intenabled = TRUE;
	 genint = TRUE;
	 break;

      case 11: /* Stop */
	 dev->select = TRUE;
	 dev->cbreak = TRUE;
	 dev->count = 0;
	 break;

      case 12: /* Step Head */
	 break;

      default: ; /* Maint. commands */
      }
#ifdef DEBUGFLOPPY
      fprintf (stderr, "\n");
#endif
      if (genint)
	 geninterrupt (dev->intlvl);
   }
   return;
}

/***********************************************************************
* fpystcr - Store CRU.
***********************************************************************/

uint16
fpystcr (Device *dev, uint16 cnt)
{
   int i;
   int genint;
   uint16 devaddr;
   uint16 dat;
#ifdef DEBUGFLOPPY
   uint8 ch;
   uint8 ch1;
#endif

   dat = 0;
   devaddr = R12;


#ifdef DEBUGFLOPPY
   fprintf (stderr, "INTMASK %X STCR %04X", GET_MASK, devaddr);
#endif

   genint = dev->intenabled;
   if (devaddr == dev->devaddr) /* Data transfer */
   {
      uint16 d;
      uint16 c;

      for (i = 0; i < devcnt; i++) /* find selected unit */
	 if (devices[i].select && devices[i].devaddr == devaddr)
	    break;
      if (i == devcnt)
      {
	 dev->info.fpyinfo.laststat = 0x4; /* Drive not ready */
	 return (dat);
      }
      dev = &devices[i];
      d = fgetc (dev->fd);
      c = fgetc (dev->fd);
      d = (d << 8) | c;
      dat = d;
#ifdef DEBUGFLOPPY
      ch = (dat >> 8) & 0xFF;
      if (!isprint (ch)) ch = '.';
      ch1 = dat & 0xFF;
      if (!isprint (ch1)) ch1 = '.';
      fprintf (stderr, " DATA %04X %c%c\n", dat, ch, ch1);
#endif
      dev->count -= 2;
      if (dev->count && genint)
	 geninterrupt (dev->intlvl);
   }

   else /* Command */
   {
      uint16 unit = 0;

      for (i = 0; i < devcnt; i++)
	 if (devices[i].select && devices[i].cdevaddr == devaddr)
	 {
	    dev = &devices[i];
	    unit = dev->unit;
#ifdef DEBUGFLOPPY
	    fprintf (stderr, " UNIT %d", unit);
#endif

	    if (dev->cbreak)
	    {
	       dev->cbreak = FALSE;
	       dev->select = FALSE;
	    }
	    break;
	 }
      if (dev->info.fpyinfo.laststat)
	 dat = dev->info.fpyinfo.laststat | 0x8001;
      else
	 dat = 0x8003 | (unit << 13);

#ifdef DEBUGFLOPPY
      fprintf (stderr, " DATA %04X\n", dat);
#endif
   }

   return (dat);
}

/***********************************************************************
* fpyboot - Boot from device.
***********************************************************************/

int
fpyboot (Device *dev)
{
   uint16 lp;
#if !defined(USEFPYROM)
   int i;
   uint16 val;
   int sect;
#endif

   run = FALSE;
   lp = LOADADDRESS;
   SET_MASK (0);

#ifdef USEFPYROM
   if (model < 10)
   {
      wpreg = GETMEM0 (0xFFFC) & 0xFFFE;
      pcreg = GETMEM0 (0xFFFE) & 0xFFFE;
   }
   else
   {
      wpreg = GETMEM0 (TPCSSTART+0xFFFC) & 0xFFFE;
      pcreg = GETMEM0 (TPCSSTART+0xFFFE) & 0xFFFE;
      PUTREG (2, LOADADDRESS);
   }
#else /* !USEFPYROM */
   if (fseek (dev->fd, 0, SEEK_SET) < 0)
   {
      sprintf (view[0], "fpyboot: seek failed: %s",
	    strerror (ERRNO));
      sprintf (view[1], "filename: %s", dev->file);
      return (-1);
   }
   sect = 0;

   do {

      for (i = 0; i < FPYSECTLEN; i+=2)
      {
	 int c;

	 if ((c = fgetc (dev->fd)) == EOF)
	 {
	    sprintf (view[0], "fpyboot: read failed: %s",
		  strerror (ERRNO));
	    sprintf (view[1], "filename: %s", dev->file);
	    return (-1);
	 }
	 val = (c & 0xFF) << 8;
	 if ((c = fgetc (dev->fd)) == EOF)
	 {
	    sprintf (view[0], "fpyboot: read failed: %s",
		  strerror (ERRNO));
	    sprintf (view[1], "filename: %s", dev->file);
	    return (-1);
	 }
	 val = val | (c & 0xFF);
	 PUTMEM0 (lp, val);
	 lp += 2;
	 if (lp >= memlen)
	 {
	    sprintf (view[0], "fpyboot: image larger than memory");
	    sprintf (view[1], "filename: %s", dev->file);
	    return (-1);
	 }
      }

      sect++;
      lp -= 2;

   } while (sect < (FPYSECTTRK * FPYTRKCYL) && (val != 0xFFFF));
#ifdef DUMPBOOT
   HEXDUMP (stderr, &memory[LOADADDRESS], lp - LOADADDRESS, 0xF1F0);
#endif

   wpreg = 0x80;
   PUTMEM0 (LOADADDRESS, (dev->unit << 10));
   pcreg = LOADADDRESS + 2;
#endif /* USEFPYROM */

   run = TRUE;

   return (0);
}

